@@ -1,13 +1,9 @@
 package Role::Tiny;
-
-sub _getglob { \*{$_[0]} }
-sub _getstash { \%{"$_[0]::"} }
-
 use strict;
-use warnings FATAL => 'all';
+use warnings;
 
-our $VERSION = '1.003003';
-$VERSION = eval $VERSION;
+our $VERSION = '2.000_009';
+$VERSION =~ tr/_//d;
 
 our %INFO;
 our %APPLIED_TO;
@@ -19,40 +15,84 @@ our @ON_ROLE_CREATE;
 
 BEGIN {
   *_WORK_AROUND_BROKEN_MODULE_STATE = "$]" < 5.009 ? sub(){1} : sub(){0};
+  *_WORK_AROUND_HINT_LEAKAGE
+    = "$]" < 5.011 && !("$]" >= 5.009004 && "$]" < 5.010001)
+      ? sub(){1} : sub(){0};
   *_MRO_MODULE = "$]" < 5.010 ? sub(){"MRO/Compat.pm"} : sub(){"mro.pm"};
 }
 
+sub _getglob { no strict 'refs'; \*{$_[0]} }
+sub _getstash { no strict 'refs'; \%{"$_[0]::"} }
+
+sub croak {
+  require Carp;
+  no warnings 'redefine';
+  *croak = \&Carp::croak;
+  goto &Carp::croak;
+}
+
 sub Role::Tiny::__GUARD__::DESTROY {
   delete $INC{$_[0]->[0]} if @{$_[0]};
 }
 
 sub _load_module {
-  (my $proto = $_[0]) =~ s/::/\//g;
-  $proto .= '.pm';
-  return 1 if $INC{$proto};
+  my ($module) = @_;
+  (my $file = "$module.pm") =~ s{::}{/}g;
+  return 1
+    if $INC{$file};
+
   # can't just ->can('can') because a sub-package Foo::Bar::Baz
   # creates a 'Baz::' key in Foo::Bar's symbol table
-  return 1 if grep !/::$/, keys %{_getstash($_[0])||{}};
+  return 1
+    if grep !/::\z/, keys %{_getstash($module)};
   my $guard = _WORK_AROUND_BROKEN_MODULE_STATE
-    && bless([ $proto ], 'Role::Tiny::__GUARD__');
-  require $proto;
+    && bless([ $file ], 'Role::Tiny::__GUARD__');
+  local %^H if _WORK_AROUND_HINT_LEAKAGE;
+  require $file;
   pop @$guard if _WORK_AROUND_BROKEN_MODULE_STATE;
   return 1;
 }
 
+# returns a hashref of subs.  keys are names, values are coderefs, scalar refs
+# for inflated constants, and undef or strings for stubs.  very similar to what
+# is # stored in the stash, but omitting GLOBs and uninflating constants and
+# stubs.
+sub _all_subs {
+  my ($package) = @_;
+  my $stash = _getstash($package);
+  return {
+    map +(exists &{"${package}::${_}"} ? ($_ => \&{"${package}::${_}"}) : ()),
+    grep !/::\z/,
+    keys %$stash
+  };
+}
+
 sub import {
   my $target = caller;
   my $me = shift;
   strict->import;
-  warnings->import(FATAL => 'all');
+  warnings->import;
+  $me->_install_subs($target);
   return if $me->is_role($target); # already exported into this package
   $INFO{$target}{is_role} = 1;
-  # get symbol table reference
-  my $stash = _getstash($target);
+
+  my $nonmethods = _all_subs($target);
+  delete @{$nonmethods}{grep /\A\(/, keys %$nonmethods};
+  $INFO{$target}{nonmethods} = $nonmethods;
+
+  # a role does itself
+  $APPLIED_TO{$target} = { $target => undef };
+  foreach my $hook (@ON_ROLE_CREATE) {
+    $hook->($target);
+  }
+}
+
+sub _install_subs {
+  my ($me, $target) = @_;
+  return if $me->is_role($target);
   # install before/after/around subs
   foreach my $type (qw(before after around)) {
     *{_getglob "${target}::${type}"} = sub {
-      require Class::Method::Modifiers;
       push @{$INFO{$target}{modifiers}||=[]}, [ $type => @_ ];
       return;
     };
@@ -65,15 +105,6 @@ sub import {
     $me->apply_roles_to_package($target, @_);
     return;
   };
-  # grab all *non-constant* (stash slot is not a scalarref) subs present
-  # in the symbol table and store their refaddrs (no need to forcibly
-  # inflate constant subs into real subs) with a map to the coderefs in
-  # case of copying or re-use
-  my @not_methods = (map { *$_{CODE}||() } grep !ref($_), values %$stash);
-  @{$INFO{$target}{not_methods}={}}{@not_methods} = @not_methods;
-  # a role does itself
-  $APPLIED_TO{$target} = { $target => undef };
-  $_->($target) for @ON_ROLE_CREATE;
 }
 
 sub role_application_steps {
@@ -85,8 +116,8 @@ sub apply_single_role_to_package {
 
   _load_module($role);
 
-  die "This is apply_role_to_package" if ref($to);
-  die "${role} is not a Role::Tiny" unless $me->is_role($role);
+  croak "This is apply_role_to_package" if ref($to);
+  croak "${role} is not a Role::Tiny" unless $me->is_role($role);
 
   foreach my $step ($me->role_application_steps) {
     $me->$step($to, $role);
@@ -101,7 +132,7 @@ sub _copy_applied_list {
 
 sub apply_roles_to_object {
   my ($me, $object, @roles) = @_;
-  die "No roles supplied!" unless @roles;
+  croak "No roles supplied!" unless @roles;
   my $class = ref($object);
   # on perl < 5.8.9, magic isn't copied to all ref copies. bless the parameter
   # directly, so at least the variable passed to us will get any magic applied
@@ -117,8 +148,11 @@ sub _composite_name {
   );
 
   if (length($new_name) > 252) {
-    $new_name = $COMPOSED{abbrev}{$new_name}
-      ||= substr($new_name, 0, 250 - length $role_suffix).'__'.$role_suffix++;
+    $new_name = $COMPOSED{abbrev}{$new_name} ||= do {
+      my $abbrev = substr $new_name, 0, 250 - length $role_suffix;
+      $abbrev =~ s/(?<!:):$//;
+      $abbrev.'__'.$role_suffix++;
+    };
   }
   return wantarray ? ($new_name, $compose_name) : $new_name;
 }
@@ -126,14 +160,13 @@ sub _composite_name {
 sub create_class_with_roles {
   my ($me, $superclass, @roles) = @_;
 
-  die "No roles supplied!" unless @roles;
+  croak "No roles supplied!" unless @roles;
 
   _load_module($superclass);
   {
     my %seen;
-    $seen{$_}++ for @roles;
-    if (my @dupes = grep $seen{$_} > 1, @roles) {
-      die "Duplicated roles: ".join(', ', @dupes);
+    if (my @dupes = grep 1 == $seen{$_}++, @roles) {
+      croak "Duplicated roles: ".join(', ', @dupes);
     }
   }
 
@@ -143,7 +176,7 @@ sub create_class_with_roles {
 
   foreach my $role (@roles) {
     _load_module($role);
-    die "${role} is not a Role::Tiny" unless $me->is_role($role);
+    croak "${role} is not a Role::Tiny" unless $me->is_role($role);
   }
 
   require(_MRO_MODULE);
@@ -155,10 +188,10 @@ sub create_class_with_roles {
       join "\n",
         map {
           "Method name conflict for '$_' between roles "
-          ."'".join(' and ', sort values %{$conflicts{$_}})."'"
+          ."'".join("' and '", sort values %{$conflicts{$_}})."'"
           .", cannot apply these simultaneously to an object."
         } keys %conflicts;
-    die $fail;
+    croak $fail;
   }
 
   my @composable = map $me->_composable_package_for($_), reverse @roles;
@@ -209,7 +242,7 @@ sub apply_roles_to_package {
           ."'".join(' and ', sort values %{$conflicts{$_}})."'"
           .", the method '$_' must be implemented by '${to}'"
         } keys %conflicts;
-    die $fail;
+    croak $fail;
   }
 
   # conflicting methods are supposed to be treated as required by the
@@ -314,7 +347,7 @@ sub _check_requires {
     if (my $to_info = $INFO{$to}) {
       push @{$to_info->{requires}||=[]}, @requires_fail;
     } else {
-      die "Can't apply ${name} to ${to} - missing ".join(', ', @requires_fail);
+      croak "Can't apply ${name} to ${to} - missing ".join(', ', @requires_fail);
     }
   }
 }
@@ -322,24 +355,41 @@ sub _check_requires {
 sub _concrete_methods_of {
   my ($me, $role) = @_;
   my $info = $INFO{$role};
-  # grab role symbol table
-  my $stash = _getstash($role);
-  # reverse so our keys become the values (captured coderefs) in case
-  # they got copied or re-used since
-  my $not_methods = { reverse %{$info->{not_methods}||{}} };
-  $info->{methods} ||= +{
-    # grab all code entries that aren't in the not_methods list
-    map {
-      my $code = *{$stash->{$_}}{CODE};
-      ( ! $code or exists $not_methods->{$code} ) ? () : ($_ => $code)
-    } grep !ref($stash->{$_}), keys %$stash
-  };
+
+  return $info->{methods}
+    if $info && $info->{methods};
+
+  my $nonmethods = ($info && $info->{nonmethods}) || {};
+
+  # this is only for backwards compatibility with older Moo, which
+  # reimplements method tracking rather than calling our method
+  my $not_methods = ($info && $info->{not_methods}) || {};
+
+  my $subs = _all_subs($role);
+  for my $sub (keys %$subs) {
+    my $code = $subs->{$sub};
+    if (
+      length ref $code && exists $not_methods->{$code}
+      or (
+        $sub !~ /\A\(/
+        and exists $nonmethods->{$sub}
+        and $code == $nonmethods->{$sub}
+      )
+    ) {
+      delete $subs->{$sub};
+    }
+  }
+
+  if ($info) {
+    $info->{methods} = $subs;
+  }
+  return $subs;
 }
 
 sub methods_provided_by {
   my ($me, $role) = @_;
-  die "${role} is not a Role::Tiny" unless $me->is_role($role);
-  (keys %{$me->_concrete_methods_of($role)}, @{$INFO{$role}->{requires}||[]});
+  croak "${role} is not a Role::Tiny" unless $me->is_role($role);
+  sort (keys %{$me->_concrete_methods_of($role)}, @{$INFO{$role}->{requires}||[]});
 }
 
 sub _install_methods {
@@ -352,15 +402,15 @@ sub _install_methods {
   # grab target symbol table
   my $stash = _getstash($to);
 
-  # determine already extant methods of target
-  my %has_methods;
-  @has_methods{grep
-    +(ref($stash->{$_}) || *{$stash->{$_}}{CODE}),
-    keys %$stash
-  } = ();
+  foreach my $i (keys %$methods) {
+    my $target = $stash->{$i};
 
-  foreach my $i (grep !exists $has_methods{$_}, keys %$methods) {
     no warnings 'once';
+    no strict 'refs';
+
+    next
+      if exists &{"${to}::${i}"};
+
     my $glob = _getglob "${to}::${i}";
     *$glob = $methods->{$i};
 
@@ -368,10 +418,10 @@ sub _install_methods {
     # and &overload::nil in the code slot.
     next
       unless $i =~ /^\(/
-        && defined &overload::nil
-        && $methods->{$i} == \&overload::nil;
+        && ((defined &overload::nil && $methods->{$i} == \&overload::nil)
+            || (defined &overload::_nil && $methods->{$i} == \&overload::_nil));
 
-    my $overload = ${ *{_getglob "${role}::${i}"}{SCALAR} };
+    my $overload = ${ _getglob "${role}::${i}" };
     next
       unless defined $overload;
 
@@ -384,10 +434,16 @@ sub _install_methods {
 sub _install_modifiers {
   my ($me, $to, $name) = @_;
   return unless my $modifiers = $INFO{$name}{modifiers};
-  if (my $info = $INFO{$to}) {
-    push @{$info->{modifiers}}, @{$modifiers||[]};
-  } else {
-    foreach my $modifier (@{$modifiers||[]}) {
+  my $info = $INFO{$to};
+  my $existing = ($info ? $info->{modifiers} : $COMPOSED{modifiers}{$to}) ||= [];
+  my @modifiers = grep {
+    my $modifier = $_;
+    !grep $_ == $modifier, @$existing;
+  } @{$modifiers||[]};
+  push @$existing, @modifiers;
+
+  if (!$info) {
+    foreach my $modifier (@modifiers) {
       $me->_install_single_modifier($to, @$modifier);
     }
   }
@@ -399,9 +455,11 @@ sub _install_single_modifier {
   my ($me, @args) = @_;
   defined($vcheck_error) or $vcheck_error = do {
     local $@;
-    eval { Class::Method::Modifiers->VERSION(1.05); 1 }
-      ? 0
-      : $@
+    eval {
+      require Class::Method::Modifiers;
+      Class::Method::Modifiers->VERSION(1.05);
+      1;
+    } ? 0 : $@;
   };
   $vcheck_error and die $vcheck_error;
   Class::Method::Modifiers::install_modifier(@args);
@@ -414,8 +472,9 @@ sub _install_does {
   # only add does() method to classes
   return if $me->is_role($to);
 
+  my $does = $me->can('does_role');
   # add does() only if they don't have one
-  *{_getglob "${to}::does"} = \&does_role unless $to->can('does');
+  *{_getglob "${to}::does"} = $does unless $to->can('does');
 
   return
     if $to->can('DOES') and $to->can('DOES') != (UNIVERSAL->can('DOES') || 0);
@@ -423,10 +482,10 @@ sub _install_does {
   my $existing = $to->can('DOES') || $to->can('isa') || $FALLBACK;
   my $new_sub = sub {
     my ($proto, $role) = @_;
-    Role::Tiny::does_role($proto, $role) or $proto->$existing($role);
+    $proto->$does($role) or $proto->$existing($role);
   };
   no warnings 'redefine';
-  *{_getglob "${to}::DOES"} = $new_sub;
+  return *{_getglob "${to}::DOES"} = $new_sub;
 }
 
 sub does_role {
@@ -440,7 +499,7 @@ sub does_role {
 
 sub is_role {
   my ($me, $role) = @_;
-  return !!($INFO{$role} && $INFO{$role}{is_role});
+  return !!($INFO{$role} && ($INFO{$role}{is_role} || $INFO{$role}{not_methods} || $INFO{$role}{nonmethods}));
 }
 
 1;
@@ -450,7 +509,7 @@ __END__
 
 =head1 NAME
 
-Role::Tiny - Roles. Like a nouvelle cuisine portion size slice of Moose.
+Role::Tiny - Roles: a nouvelle cuisine portion size slice of Moose
 
 =head1 SYNOPSIS
 
@@ -462,11 +521,11 @@ Role::Tiny - Roles. Like a nouvelle cuisine portion size slice of Moose.
 
  sub bar { ... }
 
- around baz => sub { ... }
+ around baz => sub { ... };
 
  1;
 
-else where
+elsewhere
 
  package Some::Class;
 
@@ -498,7 +557,8 @@ inheritance.  The basics of this implementation of roles is:
 =item *
 
 If a method is already defined on a class, that method will not be composed in
-from the role.
+from the role. A method inherited by a class gets overridden by the role's
+method of the same name, though.
 
 =item *
 
@@ -577,11 +637,7 @@ both L<Class::Method::Modifiers> and L<Role::Tiny>.
 =head2 Strict and Warnings
 
 In addition to importing subroutines, using C<Role::Tiny> applies L<strict> and
-L<fatal warnings|perllexwarn/Fatal Warnings> to the caller.  It's possible to
-disable these if desired:
-
- use Role::Tiny;
- use warnings NONFATAL => 'all';
+L<warnings> to the caller.
 
 =head1 SUBROUTINES
 
@@ -620,8 +676,9 @@ Composes role with package.  See also L<Role::Tiny::With>.
 
  Role::Tiny->apply_roles_to_object($foo, qw(Some::Role1 Some::Role2));
 
-Composes roles in order into object directly.  Object is reblessed into the
-resulting class.
+Composes roles in order into object directly. Object is reblessed into the
+resulting class. Note that the object's methods get overridden by the role's
+ones with the same names.
 
 =head2 create_class_with_roles
 
@@ -641,7 +698,10 @@ Returns true if the given package is a role.
 =over 4
 
 =item * On perl 5.8.8 and earlier, applying a role to an object won't apply any
-overloads from the role to all copies of the object.
+overloads from the role to other copies of the object.
+
+=item * On perl 5.16 and earlier, applying a role to a class won't apply any
+overloads from the role to any existing instances of the class.
 
 =back
 
